/*
	Basic Extensions for TI-89
		digital signal processing 
		special functions
		statistics
		variable management
		wrappers for ROM calls otherwise not available to basic
					
    Copyright (C) 2015  Ralph Schenn with support from 
    	Kevin Kofler
    	Lionel Debroux
    	Benjamin Moody

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

/* change xxx when adding functions */

#define VERSION "BE 0.5.0 \xA9 2015 Ralph Schenn rschenn\x40t-online.de"

#include <tiams.h>
#include "Bessel.h"
#include "Complex.h"
#include "ConRecEq.h"
#include "Debug.h"
#include "DSP.h"
#include "Erf.h"
#include "FFT.h"
#include "Gamma.h"
#include "Goertzel.h"
#include "Matrix.h"
#include "MovAvg.h"
#include "MultiExpr.h"
#include "Poly.h"
#include "QSort.h"
#include "Real.h"
#include "ROM.h"
#include "Search.h"
#include "Stack.h"
#include "Stat.h"
#include "String.h"
#include "Tags.h"
#include "Wrapper.h"

/* String numbers xxx */
#define H_about		0 // ok
#define H_besseli	1 // ok
#define H_besselj	2 // ok
#define H_besselk	3 // ok
#define H_bessely	4 // ok
#define H_bsearch	5 // ok
#define H_clrio		6 // ok
#define H_conreceq	7 // ok
#define H_czt		8 // 
#define H_dct		9 // 
#define H_downsamp	10 // ok
#define H_erashelp	11 // ok
#define H_erf		12 // ok
#define H_erfi		13 // ok
#define H_fdelete	14 // ok
#define H_fft		15 // ok
#define H_gamma		16 // ok
#define H_getkeypt	17 // ok
#define H_gettagst	18 // ok
#define H_goertzel	19 // ok
#define H_heapavai	20 // ok
#define H_helpmsg	21 // ok
#define H_isantsym	22 // ok
#define H_isbool	23 // ok
#define H_iscfloat	24 // ok
#define H_isconst	25 // ok
#define H_isfloat	26 // ok
#define H_isfretag	27 // ok
#define H_isindep	28 // ok
#define H_isint		29 // ok
#define H_isneg		30 // ok
#define H_isnever0	31 // ok
#define H_ispos		32 // ok
#define H_isreal	33 // ok
#define H_isstr		34 // ok
#define H_issym		35 // ok
#define H_istrue	36 // ok
#define H_isvar		37 // ok
#define H_lstcondu	38 // ok
#define H_lsteven	39 // ok
#define H_lstint	40 // ok
#define H_lstodd	41 // ok
#define H_lstrmdup	42 // ok
#define H_lstsub	43 // ok
#define H_lstuni	44 // ok
#define H_movavg	45 // ok
//#define H_nanoches	41 // 
#define H_nxtarbin	46 // ok
#define H_nxtarbre	47 // ok
#define H_pause		48 // ok
#define H_polycoef	49 // ok
#define H_polydeg	50 // ok
#define H_randseed	51 // ok
#define H_reccatse	52 // ok
//#define H_recstrse	50 // 
#define H_recvarse	53 // ok
#define H_reverse	54 // ok
#define H_rms		55 // ok
#define H_setarbin	56 // ok
#define H_setarbre	57 // ok
#define H_sorta		58 // ok
#define H_sortd		59 // ok 
#define H_ssearch	60 // ok 
#define H_stacesim	61 // ok
#define H_stacisim	62 // ok
#define H_stack		63 // ok   
#define H_supossin	64 // ok
#define H_upsamp	65 // ok
#define H_zzz		66 //  

#define H_HELP 		100

/* prototypes of extension functions xxx */
void about(void);
void besseli(void);
void besselj(void);
void besselk(void);
void bessely(void);
void bsearch(void);
void clrio(void);
void conreceq(void);	// LST
void czt(void);			// COMP LST
void dct(void);			// LST
void downsamp(void);	// LST
void erashelp(void);
void erf(void);
void erfi(void);
void fdelete(void);
void fft(void);			// COMP LST
void gamma(void);
void getkeypt(void);
void gettagst(void);
void goertzel(void);    // LST
void heapavai(void);
void helpmsg(void);
void isantsym(void);
void isbool(void);    	// LST|MAT
void iscfloat(void);    // LST|MAT
void isconst(void);
void isfloat(void);    	// LST|MAT
void isfretag(void);
void isindep(void);
void isint(void);    	// LST|MAT
void isneg(void);
void isnever0(void);
void ispos(void);
void isreal(void);
void isstr(void);		// LST|MAT
void issym(void);
void istrue(void);		// LST|MAT
void isvar(void);
void lstcondu(void);	// LST
void lsteven(void);		// LST
void lstint(void);		// LST
void lstodd(void);		// LST
void lstrmdup(void);    // LST
void lstsub(void);		// LST
void lstuni(void);		// LST
void movavg(void);		// LST
//void nanoches(void);
void nxtarbin(void);
void nxtarbre(void);
void pause(void);
void polycoef(void);
void polydeg(void);
void randseed(void);
void reccatse(void);
//void recstrse(void);
void recvarse(void);
void reverse(void);     // LST
void rms(void);    		// LST
void setarbin(void);
void setarbre(void);
void sorta(void);		// LST
void sortd(void);       // LST
void ssearch(void);
void stacesim(void);
void stacisim(void);
void stack(void);
void supossin(void);	// MAT
void upsamp(void);		// LST
void zzz(void);

/* extension table with entry for each func/prog xxx */
APP_EXTENSION const extensions[] =
{
	/* function name #, help string #, function index (to extEntries) */
	{OO_APPSTRING+H_about, OO_APPSTRING+H_HELP+H_about, H_about },
	{OO_APPSTRING+H_besseli, OO_APPSTRING+H_HELP+H_besseli, H_besseli },
	{OO_APPSTRING+H_besselj, OO_APPSTRING+H_HELP+H_besselj, H_besselj },
	{OO_APPSTRING+H_besselk, OO_APPSTRING+H_HELP+H_besselk, H_besselk },
	{OO_APPSTRING+H_bessely, OO_APPSTRING+H_HELP+H_bessely, H_bessely },
	{OO_APPSTRING+H_bsearch, OO_APPSTRING+H_HELP+H_bsearch, H_bsearch },
	{OO_APPSTRING+H_clrio, OO_APPSTRING+H_HELP+H_clrio, H_clrio },
	{OO_APPSTRING+H_conreceq, OO_APPSTRING+H_HELP+H_conreceq, H_conreceq },
	{OO_APPSTRING+H_czt, OO_APPSTRING+H_HELP+H_czt, H_czt },
	{OO_APPSTRING+H_dct, OO_APPSTRING+H_HELP+H_dct, H_dct },
	{OO_APPSTRING+H_downsamp, OO_APPSTRING+H_HELP+H_downsamp, H_downsamp },
	{OO_APPSTRING+H_erashelp, OO_APPSTRING+H_HELP+H_erashelp, H_erashelp },
	{OO_APPSTRING+H_erf, OO_APPSTRING+H_HELP+H_erf, H_erf },
	{OO_APPSTRING+H_erfi, OO_APPSTRING+H_HELP+H_erfi, H_erfi },
	//{OO_APPSTRING+H_fconv, OO_APPSTRING+H_HELP+H_fconv, H_fconv },
	{OO_APPSTRING+H_fdelete, OO_APPSTRING+H_HELP+H_fdelete, H_fdelete },
	{OO_APPSTRING+H_fft, OO_APPSTRING+H_HELP+H_fft, H_fft },
	{OO_APPSTRING+H_gamma, OO_APPSTRING+H_HELP+H_gamma, H_gamma },
	{OO_APPSTRING+H_getkeypt, OO_APPSTRING+H_HELP+H_getkeypt, H_getkeypt },
	{OO_APPSTRING+H_gettagst, OO_APPSTRING+H_HELP+H_gettagst, H_gettagst },
	{OO_APPSTRING+H_goertzel, OO_APPSTRING+H_HELP+H_goertzel, H_goertzel },
	{OO_APPSTRING+H_heapavai, OO_APPSTRING+H_HELP+H_heapavai, H_heapavai },
	{OO_APPSTRING+H_helpmsg, OO_APPSTRING+H_HELP+H_helpmsg, H_helpmsg },
	{OO_APPSTRING+H_isantsym, OO_APPSTRING+H_HELP+H_isantsym, H_isantsym },
	{OO_APPSTRING+H_isbool, OO_APPSTRING+H_HELP+H_isbool, H_isbool },
	{OO_APPSTRING+H_iscfloat, OO_APPSTRING+H_HELP+H_iscfloat, H_iscfloat },
	{OO_APPSTRING+H_isconst, OO_APPSTRING+H_HELP+H_isconst, H_isconst },
	{OO_APPSTRING+H_isfloat, OO_APPSTRING+H_HELP+H_isfloat, H_isfloat },
	{OO_APPSTRING+H_isfretag, OO_APPSTRING+H_HELP+H_isfretag, H_isfretag },
	{OO_APPSTRING+H_isindep, OO_APPSTRING+H_HELP+H_isindep, H_isindep },
	{OO_APPSTRING+H_isint, OO_APPSTRING+H_HELP+H_isint, H_isint },
	{OO_APPSTRING+H_isneg, OO_APPSTRING+H_HELP+H_isneg, H_isneg },
	{OO_APPSTRING+H_isnever0, OO_APPSTRING+H_HELP+H_isnever0, H_isnever0 },
	//{OO_APPSTRING+H_ispolvar, OO_APPSTRING+H_HELP+H_ispolvar, H_ispolvar },
	{OO_APPSTRING+H_ispos, OO_APPSTRING+H_HELP+H_ispos, H_ispos },
	{OO_APPSTRING+H_isreal, OO_APPSTRING+H_HELP+H_isreal, H_isreal },
	{OO_APPSTRING+H_isstr, OO_APPSTRING+H_HELP+H_isstr, H_isstr },
	{OO_APPSTRING+H_issym, OO_APPSTRING+H_HELP+H_issym, H_issym },
	{OO_APPSTRING+H_istrue, OO_APPSTRING+H_HELP+H_istrue, H_istrue },
	{OO_APPSTRING+H_isvar, OO_APPSTRING+H_HELP+H_isvar, H_isvar },
	{OO_APPSTRING+H_lstcondu, OO_APPSTRING+H_HELP+H_lstcondu, H_lstcondu },
	{OO_APPSTRING+H_lsteven, OO_APPSTRING+H_HELP+H_lsteven, H_lsteven },
	{OO_APPSTRING+H_lstint, OO_APPSTRING+H_HELP+H_lstint, H_lstint },
	{OO_APPSTRING+H_lstodd, OO_APPSTRING+H_HELP+H_lstodd, H_lstodd },
	{OO_APPSTRING+H_lstrmdup, OO_APPSTRING+H_HELP+H_lstrmdup, H_lstrmdup },
	{OO_APPSTRING+H_lstsub, OO_APPSTRING+H_HELP+H_lstsub, H_lstsub },
	{OO_APPSTRING+H_lstuni, OO_APPSTRING+H_HELP+H_lstuni, H_lstuni },
	{OO_APPSTRING+H_movavg, OO_APPSTRING+H_HELP+H_movavg, H_movavg },
	//{OO_APPSTRING+H_nanoches, OO_APPSTRING+H_HELP+H_nanoches, H_nanoches },
	{OO_APPSTRING+H_nxtarbin, OO_APPSTRING+H_HELP+H_nxtarbin, H_nxtarbin },
	{OO_APPSTRING+H_nxtarbre, OO_APPSTRING+H_HELP+H_nxtarbre, H_nxtarbre },
	{OO_APPSTRING+H_pause, OO_APPSTRING+H_HELP+H_pause, H_pause },
	{OO_APPSTRING+H_polycoef, OO_APPSTRING+H_HELP+H_polycoef, H_polycoef },
	{OO_APPSTRING+H_polydeg, OO_APPSTRING+H_HELP+H_polydeg, H_polydeg },
	{OO_APPSTRING+H_randseed, OO_APPSTRING+H_HELP+H_randseed, H_randseed },
	{OO_APPSTRING+H_reccatse, OO_APPSTRING+H_HELP+H_reccatse, H_reccatse },
	//{OO_APPSTRING+H_recstrse, OO_APPSTRING+H_HELP+H_recstrse, H_recstrse },
	{OO_APPSTRING+H_recvarse, OO_APPSTRING+H_HELP+H_recvarse, H_recvarse },
	{OO_APPSTRING+H_reverse, OO_APPSTRING+H_HELP+H_reverse, H_reverse },
	{OO_APPSTRING+H_rms, OO_APPSTRING+H_HELP+H_rms, H_rms },
	{OO_APPSTRING+H_setarbin, OO_APPSTRING+H_HELP+H_setarbin, H_setarbin },
	{OO_APPSTRING+H_setarbre, OO_APPSTRING+H_HELP+H_setarbre, H_setarbre },
	{OO_APPSTRING+H_sorta, OO_APPSTRING+H_HELP+H_sorta, H_sorta },
	{OO_APPSTRING+H_sortd, OO_APPSTRING+H_HELP+H_sortd, H_sortd },
	{OO_APPSTRING+H_ssearch, OO_APPSTRING+H_HELP+H_ssearch, H_ssearch },
	{OO_APPSTRING+H_stacesim, OO_APPSTRING+H_HELP+H_stacesim, H_stacesim },
	{OO_APPSTRING+H_stacisim, OO_APPSTRING+H_HELP+H_stacisim, H_stacisim },
	{OO_APPSTRING+H_stack, OO_APPSTRING+H_HELP+H_stack, H_stack },
	{OO_APPSTRING+H_supossin, OO_APPSTRING+H_HELP+H_supossin, H_supossin },
	{OO_APPSTRING+H_upsamp, OO_APPSTRING+H_HELP+H_upsamp, H_upsamp }
	//{OO_APPSTRING+H_zzz, OO_APPSTRING+H_HELP+H_zzz, H_zzz }
};

/* cross-ref of func num from extensions table with func address 
(c func name and basic func name may be differennt) xxx */
APP_EXT_ENTRY const extEntries[] =
{           
	/* c name of func, type of ext (APP_EXT_FUNCTION or APP_EXT_PROGRAM) */
	{about, APP_EXT_FUNCTION},
	{besseli, APP_EXT_FUNCTION},
	{besselj, APP_EXT_FUNCTION},
	{besselk, APP_EXT_FUNCTION},
	{bessely, APP_EXT_FUNCTION},
	{bsearch, APP_EXT_FUNCTION},
	{clrio, APP_EXT_PROGRAM},
	{conreceq, APP_EXT_FUNCTION},
	{czt, APP_EXT_FUNCTION},
	{dct, APP_EXT_FUNCTION},
	{downsamp, APP_EXT_FUNCTION},
	{erashelp, APP_EXT_PROGRAM},
	{erf, APP_EXT_FUNCTION},
	{erfi, APP_EXT_FUNCTION},
	//{fconv, APP_EXT_FUNCTION},
	{fdelete, APP_EXT_PROGRAM},
	{fft, APP_EXT_FUNCTION},
	{gamma, APP_EXT_FUNCTION},
	{getkeypt, APP_EXT_FUNCTION},
	{gettagst, APP_EXT_FUNCTION},
	{goertzel, APP_EXT_FUNCTION},
	{heapavai, APP_EXT_FUNCTION},
	{helpmsg, APP_EXT_PROGRAM},
	{isantsym, APP_EXT_FUNCTION},
	{isbool, APP_EXT_FUNCTION},
	{iscfloat, APP_EXT_FUNCTION},
	{isconst, APP_EXT_FUNCTION},
	{isfloat, APP_EXT_FUNCTION},
	{isfretag, APP_EXT_FUNCTION},
	{isindep, APP_EXT_FUNCTION},
	{isint, APP_EXT_FUNCTION},
	{isneg, APP_EXT_FUNCTION},
	{isnever0, APP_EXT_FUNCTION},
	//{ispolvar, APP_EXT_FUNCTION},
	{ispos, APP_EXT_FUNCTION},
	{isreal, APP_EXT_FUNCTION},
	{isstr, APP_EXT_FUNCTION},
	{issym, APP_EXT_FUNCTION},
	{istrue, APP_EXT_FUNCTION},
	{isvar, APP_EXT_FUNCTION},
	{lstcondu, APP_EXT_FUNCTION},
	{lsteven, APP_EXT_FUNCTION},
	{lstint, APP_EXT_FUNCTION},
	{lstodd, APP_EXT_FUNCTION},
	{lstrmdup, APP_EXT_FUNCTION},
	{lstsub, APP_EXT_FUNCTION},
	{lstuni, APP_EXT_FUNCTION},
	{movavg, APP_EXT_FUNCTION},
	//{nanoches, APP_EXT_PROGRAM},
	{nxtarbin, APP_EXT_FUNCTION},
	{nxtarbre, APP_EXT_FUNCTION},
	{pause, APP_EXT_PROGRAM},
	{polycoef, APP_EXT_FUNCTION},
	{polydeg, APP_EXT_FUNCTION},
	{randseed, APP_EXT_PROGRAM},
	{reccatse, APP_EXT_FUNCTION},
	//{recstrse, APP_EXT_FUNCTION},
	{recvarse, APP_EXT_FUNCTION},
	{reverse, APP_EXT_FUNCTION},
	{rms, APP_EXT_FUNCTION},
	{setarbin, APP_EXT_PROGRAM},
	{setarbre, APP_EXT_PROGRAM},
	{sorta, APP_EXT_PROGRAM},
	{sortd, APP_EXT_PROGRAM},
	{ssearch, APP_EXT_FUNCTION},
	{stacesim, APP_EXT_FUNCTION},
	{stacisim, APP_EXT_FUNCTION},
	{stack, APP_EXT_FUNCTION},
	{supossin, APP_EXT_FUNCTION},
	{upsamp, APP_EXT_FUNCTION}
	//{zzz, APP_EXT_FUNCTION}
};

/* frameName, parent, prototype, firstAttr, count (attr/meth selectors) 
cutis.movavg to call func from basic xxx */
FRAME(beFrame, OO_SYSTEM_FRAME, 0, OO_APP_FLAGS, 138)   
/* APP_NONE, APP_INTERACTIVE, APP_CON, APP_ACCESS_SYSVARS, APP_BACKGROUND */
ATTR(OO_APP_FLAGS, APP_NONE) 
ATTR(OO_APP_NAME, "Basic Ext") /* name for apps menu (interactive apps only) */
ATTR(OO_APP_TOK_NAME, "be") /* basic app name */
ATTR(OO_APP_EXT_COUNT, 66) /* # of exported extension funcs/progs xxx */
ATTR(OO_APP_EXTENSIONS, extensions) /* address of extensions table */
ATTR(OO_APP_EXT_ENTRIES, extEntries) /* address of ext entries table */

 /* help string xxx */
ATTR(OO_APPSTRING+H_about, "about")
ATTR(OO_APPSTRING+H_besseli, "besseli")
ATTR(OO_APPSTRING+H_besselj, "besselj")
ATTR(OO_APPSTRING+H_besselk, "besselk")
ATTR(OO_APPSTRING+H_bessely, "bessely")
ATTR(OO_APPSTRING+H_bsearch, "bsearch")       
ATTR(OO_APPSTRING+H_clrio, "clrio")
ATTR(OO_APPSTRING+H_conreceq, "conreceq")
ATTR(OO_APPSTRING+H_czt, "czt")
ATTR(OO_APPSTRING+H_dct, "dct")
ATTR(OO_APPSTRING+H_downsamp, "downsamp")
ATTR(OO_APPSTRING+H_erashelp, "erashelp")
ATTR(OO_APPSTRING+H_erf, "erf")
ATTR(OO_APPSTRING+H_erfi, "erfi")
//ATTR(OO_APPSTRING+H_fconv, "fconv")
ATTR(OO_APPSTRING+H_fdelete, "fdelete")
ATTR(OO_APPSTRING+H_fft, "fft")
ATTR(OO_APPSTRING+H_gamma, "gamma")
ATTR(OO_APPSTRING+H_getkeypt, "getkeypt")       
ATTR(OO_APPSTRING+H_gettagst, "gettagst")       
ATTR(OO_APPSTRING+H_goertzel, "goertzel")
ATTR(OO_APPSTRING+H_heapavai, "heapavai")
ATTR(OO_APPSTRING+H_helpmsg, "helpmsg")
ATTR(OO_APPSTRING+H_isantsym, "isantsym")
ATTR(OO_APPSTRING+H_isbool, "isbool")
ATTR(OO_APPSTRING+H_iscfloat, "iscfloat")
ATTR(OO_APPSTRING+H_isconst, "isconst")
ATTR(OO_APPSTRING+H_isfloat, "isfloat")
ATTR(OO_APPSTRING+H_isfretag, "isfretag")
ATTR(OO_APPSTRING+H_isindep, "isindep")
ATTR(OO_APPSTRING+H_isint, "isint")
ATTR(OO_APPSTRING+H_isneg, "isneg")
ATTR(OO_APPSTRING+H_isnever0, "isnever0")
//ATTR(OO_APPSTRING+H_ispolvar, "ispolvar")
ATTR(OO_APPSTRING+H_ispos, "ispos")
ATTR(OO_APPSTRING+H_isreal, "isreal")
ATTR(OO_APPSTRING+H_isstr, "isstr")
ATTR(OO_APPSTRING+H_issym, "issym")
ATTR(OO_APPSTRING+H_istrue, "istrue")
ATTR(OO_APPSTRING+H_isvar, "isvar")
ATTR(OO_APPSTRING+H_lstcondu, "lstcondu")
ATTR(OO_APPSTRING+H_lsteven, "lsteven")
ATTR(OO_APPSTRING+H_lstint, "lstint")
ATTR(OO_APPSTRING+H_lstodd, "lstodd")
ATTR(OO_APPSTRING+H_lstrmdup, "lstrmdup")       
ATTR(OO_APPSTRING+H_lstsub, "lstsub")
ATTR(OO_APPSTRING+H_lstuni, "lstuni")
ATTR(OO_APPSTRING+H_movavg, "movavg")       
//ATTR(OO_APPSTRING+H_nanoches, "nanoches")       
ATTR(OO_APPSTRING+H_nxtarbin, "nxtarbin")       
ATTR(OO_APPSTRING+H_nxtarbre, "nxtarbre")       
ATTR(OO_APPSTRING+H_pause, "pause")       
ATTR(OO_APPSTRING+H_polycoef, "polycoef")       
ATTR(OO_APPSTRING+H_polydeg, "polydeg")       
ATTR(OO_APPSTRING+H_randseed, "randseed")       
ATTR(OO_APPSTRING+H_reccatse, "reccatse")       
//ATTR(OO_APPSTRING+H_recstrse, "recstrse")       
ATTR(OO_APPSTRING+H_recvarse, "recvarse")       
ATTR(OO_APPSTRING+H_reverse, "reverse")       
ATTR(OO_APPSTRING+H_rms, "rms")       
ATTR(OO_APPSTRING+H_setarbin, "setarbin")       
ATTR(OO_APPSTRING+H_setarbre, "setarbre")       
ATTR(OO_APPSTRING+H_sorta, "sorta")       
ATTR(OO_APPSTRING+H_sortd, "sortd")       
ATTR(OO_APPSTRING+H_ssearch, "ssearch")       
ATTR(OO_APPSTRING+H_stacesim, "stacesim")       
ATTR(OO_APPSTRING+H_stacisim, "stacisim")       
ATTR(OO_APPSTRING+H_stack, "stack")       
ATTR(OO_APPSTRING+H_supossin, "supossin")       
ATTR(OO_APPSTRING+H_upsamp, "upsamp")       
//ATTR(OO_APPSTRING+H_zzz, "zzz")       

 /* help string xxx */
ATTR(OO_APPSTRING+H_HELP+H_about, "NO PAR")
ATTR(OO_APPSTRING+H_HELP+H_besseli, "ORD, ARG")
ATTR(OO_APPSTRING+H_HELP+H_besselj, "ORD, ARG")
ATTR(OO_APPSTRING+H_HELP+H_besselk, "ORD, ARG")
ATTR(OO_APPSTRING+H_HELP+H_bessely, "ORD, ARG")
ATTR(OO_APPSTRING+H_HELP+H_bsearch, "LST, KEY")
ATTR(OO_APPSTRING+H_HELP+H_clrio, "NO PAR")
ATTR(OO_APPSTRING+H_HELP+H_conreceq, "XLST, ALST[, BLST]")
ATTR(OO_APPSTRING+H_HELP+H_czt, "LST, LENOUT, RATIO PT, START PT")
ATTR(OO_APPSTRING+H_HELP+H_dct, "LST, 1|-1")
ATTR(OO_APPSTRING+H_HELP+H_downsamp, "LST, FACT")
ATTR(OO_APPSTRING+H_HELP+H_erashelp, "NO PAR")
ATTR(OO_APPSTRING+H_HELP+H_erf, "ARG")
ATTR(OO_APPSTRING+H_HELP+H_erfi, "ARG")
//ATTR(OO_APPSTRING+H_HELP+H_fconv, "XLST, HLST")
ATTR(OO_APPSTRING+H_HELP+H_fdelete, "VAR")
ATTR(OO_APPSTRING+H_HELP+H_fft, "CLST, -1|1[, LEN]")
ATTR(OO_APPSTRING+H_HELP+H_gamma, "ARG")
ATTR(OO_APPSTRING+H_HELP+H_getkeypt, "TAG, EXTTAG")
ATTR(OO_APPSTRING+H_HELP+H_gettagst, "TAG")
ATTR(OO_APPSTRING+H_HELP+H_goertzel, "LST, FRESAM, FRE")
ATTR(OO_APPSTRING+H_HELP+H_heapavai, "NO PAR")
ATTR(OO_APPSTRING+H_HELP+H_helpmsg, "STR")
ATTR(OO_APPSTRING+H_HELP+H_isantsym, "EXPR, VAR")
ATTR(OO_APPSTRING+H_HELP+H_isbool, "EXPR")
ATTR(OO_APPSTRING+H_HELP+H_iscfloat, "EXPR")
ATTR(OO_APPSTRING+H_HELP+H_isconst, "EXPR")
ATTR(OO_APPSTRING+H_HELP+H_isfloat, "EXPR")
ATTR(OO_APPSTRING+H_HELP+H_isfretag, "EXPR, TAGNUM")
ATTR(OO_APPSTRING+H_HELP+H_isindep, "EXPR, VAR")
ATTR(OO_APPSTRING+H_HELP+H_isint, "EXPR")
ATTR(OO_APPSTRING+H_HELP+H_isneg, "EXPR")
ATTR(OO_APPSTRING+H_HELP+H_isnever0, "EXPR")
//ATTR(OO_APPSTRING+H_HELP+H_ispolvar, "EXPR, VAR")
ATTR(OO_APPSTRING+H_HELP+H_ispos, "EXPR")
ATTR(OO_APPSTRING+H_HELP+H_isreal, "EXPR")
ATTR(OO_APPSTRING+H_HELP+H_isstr, "EXPR")
ATTR(OO_APPSTRING+H_HELP+H_issym, "EXPR, VAR")
ATTR(OO_APPSTRING+H_HELP+H_istrue, "EXPR")
ATTR(OO_APPSTRING+H_HELP+H_isvar, "EXPR")
ATTR(OO_APPSTRING+H_HELP+H_lstcondu, "LST")
ATTR(OO_APPSTRING+H_HELP+H_lsteven, "LST")
ATTR(OO_APPSTRING+H_HELP+H_lstint, "LST, LST")
ATTR(OO_APPSTRING+H_HELP+H_lstodd, "LST")
ATTR(OO_APPSTRING+H_HELP+H_lstrmdup, "LST")
ATTR(OO_APPSTRING+H_HELP+H_lstsub, "LST, LST")
ATTR(OO_APPSTRING+H_HELP+H_lstuni, "LST, LST")
ATTR(OO_APPSTRING+H_HELP+H_movavg, "LST, LEN")
//ATTR(OO_APPSTRING+H_HELP+H_nanoches, "NO PAR")
ATTR(OO_APPSTRING+H_HELP+H_nxtarbin, "NO PAR")
ATTR(OO_APPSTRING+H_HELP+H_nxtarbre, "NO PAR")
ATTR(OO_APPSTRING+H_HELP+H_pause, "EXPR")
ATTR(OO_APPSTRING+H_HELP+H_polycoef, "EXPR, VAR")
ATTR(OO_APPSTRING+H_HELP+H_polydeg, "EXPR, VAR")
ATTR(OO_APPSTRING+H_HELP+H_randseed, "EXPR")
ATTR(OO_APPSTRING+H_HELP+H_reccatse, "STR")
//ATTR(OO_APPSTRING+H_HELP+H_recstrse, "[FOLD, ]STR")
ATTR(OO_APPSTRING+H_HELP+H_recvarse, "[FOLD, ]VAR")
ATTR(OO_APPSTRING+H_HELP+H_reverse, "LST|MAT")
ATTR(OO_APPSTRING+H_HELP+H_rms, "LST")
ATTR(OO_APPSTRING+H_HELP+H_setarbin, "NUM")
ATTR(OO_APPSTRING+H_HELP+H_setarbre, "NUM")
ATTR(OO_APPSTRING+H_HELP+H_sorta, "LST1,...")
ATTR(OO_APPSTRING+H_HELP+H_sortd, "LST1,...")
ATTR(OO_APPSTRING+H_HELP+H_ssearch, "LST, KEY")
ATTR(OO_APPSTRING+H_HELP+H_stacesim, "EXPR")
ATTR(OO_APPSTRING+H_HELP+H_stacisim, "EXPR")
ATTR(OO_APPSTRING+H_HELP+H_stack, "EXPR1,...")
ATTR(OO_APPSTRING+H_HELP+H_supossin, "FRESAM, LEN, [FRE, AMP, ANG;...]")
ATTR(OO_APPSTRING+H_HELP+H_upsamp, "LST, FACT")
//ATTR(OO_APPSTRING+H_HELP+H_zzz, "FOR TEST ONLY")
ENDFRAME

/* pointer to FRAME
(declared before any other init vars to become 1. var in .data section */
pFrame BeFrame = (pFrame)&beFrame; 

//static Boolean flag0;

//////////////////////////////////////////////////////////////////////////////////////////////////////////////

/* about string */
void about(void)
{            
	push_zstr (VERSION);	
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////

/* modified Bessel function of first kind of integer order of real argument
Example: besseli(ord, arg) */
void besseli(void)
{            
	BessWrap (&BessI);
	//RealFunc (&push_besseli); 	
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////

/* Bessel function of first kind of integer order of real argument
Example: besselj(ord, arg) */
void besselj(void)
{       
	BessWrap (&BessJ);	
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////

/* modified Bessel function of third kind of integer order of real argument
Example: besselk(ord, arg) */
void besselk(void)
{       
	BessWrap (&BessK);	
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////

/* Bessel function of second kind of integer order of real argument
Example: bessely(ord, arg) */
void bessely(void)
{       
	BessWrap (&BessY);	
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////

/* binary search */
void bsearch(void)
{  
    Access_AMS_Global_Variables;

	// check number of parameters
	CheckNumArgs (2);
                	        
	// sort float list
	push_ushort_to_integer (BinSearch (top_estack, next_expression_index (top_estack)));
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////

/* clear I/O screen */
void clrio(void)
{	
	cmd_clrio ();
	
	#ifdef DEBUG
		PushStackAsMat ();
	#endif             

}
     
//////////////////////////////////////////////////////////////////////////////////////////////////////////////

/* downsample the array */
void downsamp(void)
{
	LstIntFunc (&Downsamp);
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////

/* erase help message */
void erashelp(void)
{
	ST_eraseHelp (); 
	
	#ifdef DEBUG
		PushStackAsMat ();
	#endif             
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////

/* inverse error function */
void erf(void)
{
	RealFunc (&push_erf);
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////

/* inverse error function */
void erfi(void)
{
	RealFunc (&push_erfi);
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////

/* delete file */
void fdelete(void)
{
    StrFunc ((const char * (*)(const char *))&fdeletew);
}  

//////////////////////////////////////////////////////////////////////////////////////////////////////////////

/* gamma function */
void gamma(void)
{
	RealFunc (&push_gamma);
}  

//////////////////////////////////////////////////////////////////////////////////////////////////////////////


/* get name of primary or extended tag 
for primary tag second arg = 0 
for secondary tag primary arg = 227 */
void getkeypt(void)
{
    Access_AMS_Global_Variables;  

	Quantum prim, sec; 
	
	CheckNumArgs (2);
	
	// get primary tag
	prim = estack_to_ushort_simp (top_estack);
	
	// get secondary tag
	sec = estack_to_ushort_simp (next_expression_index (top_estack));
	
	// get pointer to catalog string
	push_zstr ((char *)get_key_ptr (prim, sec)); 
	
	#ifdef DEBUG
		PushStackAsMat ();
	#endif             
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////

/* get string for tag */
void gettagst(void)
{
	Access_AMS_Global_Variables;  
	
	char name[SYM_LEN + 1]; 
	        
	CheckNumArgs (1);
		
	GetTagStrW(estack_to_ushort_simp (top_estack), name);

	push_zstr (name);
	
	#ifdef DEBUG
		PushStackAsMat ();
	#endif             	
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////

/* free heap space */
void heapavai(void)
{
	push_long_to_integer (HeapAvail ());

	#ifdef DEBUG
		PushStackAsMat ();
	#endif             	
}


//////////////////////////////////////////////////////////////////////////////////////////////////////////////

/* display help message */
void helpmsg(void)
{
	StrFunc ((const char * (*)(const char *))&sthelpmsgw);
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////

/* check if expression is constant */
void isconst(void) // 
{
	IsExpr(&isconstantw);
}   

//////////////////////////////////////////////////////////////////////////////////////////////////////////////

/* check if expression is antisymmetric */
void isantsym(void) // 
{
	ExprVar (&isantisymmetricw, FALSE, TRUE);
}   

//////////////////////////////////////////////////////////////////////////////////////////////////////////////

/* is float aggreagte */
void isfloat(void)
{
	Access_AMS_Global_Variables;
	
	push_bool (IsTypeAggregate (top_estack, &IsFloat));
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////

/* check if expression is free of tag */
void isfretag(void) // 
{
	Access_AMS_Global_Variables;          
	unsigned short q;
    Boolean b;
    
	// check number of arguments
	CheckNumArgs (2);
	
	q = estack_to_ushort_simp (next_expression_index (top_estack));
	
	if (q <= UCHAR_MAX) {
		push_simplify_statements (top_estack); 
			
		b = is_free_of_tag(top_estack, q);
		
		// remove the args from stack
		delete_between (next_expression_index (top_estack), top_estack);	
	
		// push result to estack
		push_quantum (bool_to_tag (b));
	} else {
		ER_throwVar (ER_DOMAIN);
	}
	
	#ifdef DEBUG
		PushStackAsMat ();
	#endif               
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////

/* check if expression is independant of variable or expression */
void isindep(void) //
{
	ExprVar (&isindependentofw, FALSE, TRUE);
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////

void isint(void)
{
	Access_AMS_Global_Variables;
	
	push_bool (IsTypeAggregate (top_estack, &IsInt));
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////

void isbool(void)
{
	Access_AMS_Global_Variables;
	
	push_bool (IsTypeAggregate (top_estack, &IsBool));
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////

void iscfloat(void)
{
	Access_AMS_Global_Variables;
	
	push_bool (IsTypeAggregate (top_estack, &IsCFloat));
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////

/* check if expression is negative */
void isneg(void) // 
{
	IsExpr(&isnegativew);
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////

/* check if expression is never zero */
void isnever0(void) // 
{
    IsExpr (&isnever0w);
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////

/* check if expression is polynomial in varaiable */
/*void ispolvar(void) // 
{
	ExprVar (&is_polynomial_in_var_or_kern, FALSE, TRUE);
}*/

//////////////////////////////////////////////////////////////////////////////////////////////////////////////

/* check if expression is positive */
void ispos(void) // 
{
	IsExpr (&ispositivew);
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////

/* check if expression is real */
void isreal(void) // 
{
	IsExpr(&isrealw);
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////

/* is string aggrgate */
void isstr(void)
{
	Access_AMS_Global_Variables;
	
	push_bool (IsTypeAggregate (top_estack, &IsString));
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////

/* check if expression is symmatric */
void issym(void) // 
{
	ExprVar(&issymmetricw, FALSE, TRUE);
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////

/* scalar, list or matrix is true */
void istrue(void) // 
{
	Access_AMS_Global_Variables;
	
	push_bool (IsTypeAggregate(top_estack, &IsTrue));
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////

/* check if expression is a variable */
void isvar(void) // 
{
	IsExpr(&isvariablew);
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////

/* list contains duplicates */
void lstcondu(void)
{
	Access_AMS_Global_Variables;
	
	EStackIndex argptr = top_estack - 1;
	Boolean dup = FALSE;
	
	CheckNumArgs (1);
	
	if (!is_list (top_estack)) {
		ER_throwVar (ER_ARGUMENT_MUST_BE_LIST);
	}
	
	while (ESTACK (argptr) != END_TAG) {
		if (!compare_expressions (argptr, next_expression_index (argptr))) {
			dup = TRUE;
			break;
		}
		
		argptr = next_expression_index (argptr);
	}
	
	push_bool (dup);
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////

/* even elements of list */
void lsteven(void)
{   
    LstOddEven (FALSE);
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////

/* intersection of lists */
void lstint(void)
{
	LstIntSub (FALSE); 
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////

/* odd elements of list */
void lstodd(void)
{   
    LstOddEven (TRUE);
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////

/* only push elements that are different from prevous element */
static void RmDup(EStackIndex i)
{
	// not a duplicate
	if (compare_expressions (i, next_expression_index (i)) != 0) {
		// add to result
		push_expression (i);
	}
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////

/* remove duplicates from list 
needs sort function */
void lstrmdup(void)
{
	MapTail (&RmDup);
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////

/* subtract elements of second list from first list */
void lstsub(void)
{
	LstIntSub (TRUE);
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////

/* union of lists 
map_tail is not suitable here because callback function also needs pointer to other list */
void lstuni(void)
{
	Access_AMS_Global_Variables;
	
	EStackIndex lst1 = top_estack, 
		lst2,
		lst1s,
		lst2s,
		lst2so;
	
	CheckNumArgs (2);
	
	lst2 = next_expression_index (top_estack);
	
	push_simplify_statements (lst2);
	push_simplify_statements (lst1);
		
	lst1s = top_estack;
	lst2s = next_expression_index (top_estack);
	
	if (!is_list (lst1s) || !is_list (lst2s)) {
		ER_throwVar (ER_ARGUMENT_MUST_BE_LIST);
	}
	
	// first list is copied to union
	push_expression (lst1s);
	
	// overwrite the LIST_TAG with the next push_expression
	top_estack--; 
	
	// save index for later deletion of list
	lst2so = lst2s;
	
	// start with first element of lst2
	lst2s--;
	
	// loop over second list
	while (ESTACK (lst2s) != END_TAG) {
		// element of second list not found in first list
		if (BinSearch (lst1s, lst2s) == 0) {
			// add element of second list to union
			push_expression (lst2s);
		}
		
		// next element of second list
		lst2s = next_expression_index (lst2s);
	}

	// terminate the union
	push_quantum (LIST_TAG); 
	
	// remove simplified lists
	delete_between (lst2so, lst1); 
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////

/* moving average */
void movavg(void)
{
	LstIntFunc(&MovAvg);
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////

/* next arbitrary int */
void nxtarbin(void)
{
	push_next_arb_int ();
	
	#ifdef DEBUG
		PushStackAsMat ();
	#endif             
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////

/* next arbitrary real */
void nxtarbre(void)
{
	push_next_arb_real ();
	
	#ifdef DEBUG
		PushStackAsMat ();
	#endif             
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////

/* display args on IO screen and halt */
void pause(void) // ok
{	
	ExecFuncArgs (&cmdpause, TRUE, FALSE);	
	cmd_disphome ();
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////

/* degree of polynomial */
void polydeg(void) // 
{	
	// result is scalar and not bool
	// cast to function returning bool even when this function returns void
	ExprVar ((Boolean (*)(EStackIndex, EStackIndex))&pushpolydeginvarorkernelw, FALSE, FALSE);
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////

/* coefficients of polynomial */
void polycoef(void)
{
	// result is list and not bool
	// cast to function returning bool even when this function returns void
	ExprVar ((Boolean (*)(EStackIndex, EStackIndex))&push_coef_list, TRUE, FALSE);
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////

/* seed the random number generator */
void randseed(void)
{ 
	Access_AMS_Global_Variables; 
	
	// check number of arguments
	CheckNumArgs (1);
	
	// simplify to internal form
	push_simplify_statements (top_estack);
	
	// call function
	cmdrandseed (top_estack);
	
	// delete the internal simplified function arg
	delete_expression (top_estack);
	
	#ifdef DEBUG
		PushStackAsMat ();
	#endif             
}
	        
//////////////////////////////////////////////////////////////////////////////////////////////////////////////

/* recursive string search */
/*void recstrse(void)
{
	RecSearch (&RecStrSe);
}*/
	        
//////////////////////////////////////////////////////////////////////////////////////////////////////////////

/* recursive variable search */
void recvarse(void)
{	
	RecSearch (&RecVarSe);
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////

/* reverse list */
void reverse(void)
{
    Access_AMS_Global_Variables; 
    
    EStackIndex old_top; 
	    
    CheckNumArgs (1);

	// variable may have been passed instead of list
	push_simplify_statements (top_estack);
	
	// wrong arg type
	if (!is_list (top_estack)) {
		ER_throwVar (ER_ARGUMENT_MUST_BE_LIST);		
	}	
    
	old_top = top_estack;
	
	// push reversed simplified list
	push_reversed_tail (top_estack - 1); 
		
	// beginning of list
	push_quantum (LIST_TAG);	

	// delete simplified list in original order
	delete_expression (old_top);
	
	#ifdef DEBUG
		PushStackAsMat ();
	#endif             
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////

/* root mean square */
void rms(void)
{
    Access_AMS_Global_Variables;
    
    EStackIndex argptr;    
    unsigned short n = 0;
    double s = 0., t;
    
	// check number of parameters
	CheckNumArgs (1);
    
    push_simplify_statements (top_estack);
    
    argptr = top_estack; 
    
    if (!is_list (argptr)) {
    	ER_throwVar (ER_ARGUMENT_MUST_BE_LIST);
    }
    
    argptr--;
    
    while (ESTACK (argptr) != END_TAG) {
    	t = estack_number_to_Float_inf (argptr);
    	s += t * t;
    	
    	argptr = next_expression_index (argptr);
    	n++;
    }
    
    delete_expression (top_estack);
    
    push_Float (sqrt (s / n));            
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////

/* set arbitary integer */
void setarbin(void)
{  
	SetArb (TRUE);
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////

/* set arbitary real */
void setarbre(void)
{
	SetArb (FALSE);
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////

/* sort ascending */
void sorta(void) // ok
{    
	ExecFuncArgs (&cmdsorta, FALSE, TRUE);	
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////

/* sort descending */
void sortd(void) // ok
{
	ExecFuncArgs (&cmdsortd, FALSE, TRUE);	
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////

/* sequential search 
pushes found indices in descending order 
cant use map_tail because compare_expressions needs two arguments */
void ssearch(void)
{
	Access_AMS_Global_Variables;
	
	EStackIndex argptr = top_estack, key;
	unsigned short i = 1;
	
	CheckNumArgs (2);
		
	if (!is_list (top_estack)) {
		ER_throwVar (ER_ARGUMENT_MUST_BE_LIST);
	}
	
	key = next_expression_index (top_estack);

	push_quantum (END_TAG);

	argptr--;
	
	while (ESTACK (argptr) != END_TAG) {
		// element matches
		if (!compare_expressions (argptr, key)) {
			// push index of matiching elment
			push_ushort_to_integer (i);			
		}
		
		argptr = next_expression_index (argptr);
		i++;			
	}
				
	push_quantum (LIST_TAG);
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////

/* show tags of arga on estack (top to bottom) */
void stack(void) // 
{
	PushStackAsMat ();
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////

/* show tags of external simplified arg on estack (top to bottom) */
void stacesim(void) // 
{
    Access_AMS_Global_Variables;  
    
	CheckNumArgs (1);

	push_simplify_statements (top_estack);
		
	// push contents of stack as mattrix to stack
	PushStackAsMat ();
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////

/* show tags of internal simplified arg on estack (top to bottom) */
void stacisim(void) // 
{
    Access_AMS_Global_Variables;  
    
	CheckNumArgs (1);

	push_internal_simplify (top_estack);
		
	// push contents of stack as mattrix to stack
	PushStackAsMat ();
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////

/* downsample the array */
void upsamp(void)
{
	LstIntFunc (&Upsamp);
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////

/* sort seems to be ok
with floats it works while it fails with expressions */
/*void zzz(void)
{
}*/
