/*
		Convolution or Recursion Equation
		
    Copyright (C) 2015  Ralph Schenn

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <tiams.h>
#include "ConRecEq.h"
#include "Stack.h"

/* Convolution or Recursive Equation for TI-89
pointer to lists of real numbers on estack 
two list=convoltion 
three lists=recursion equation */
void conreceq(void)
{
	Access_AMS_Global_Variables; 
	EStackIndex argptr = top_estack, topPtr;
	
	// variables referenced in ONERR or FINALLY should be volatile
	volatile HANDLE xHdl;
	volatile HANDLE aHdl;
	volatile HANDLE bHdl;
	volatile HANDLE yHdl;

	double *xPtr;
	double *aPtr;
	double *bPtr = NULL;
	double *yPtr;

	// false=convolution, true=recursion
	Boolean recEq;
	
	unsigned short lenX;
	unsigned short lenA;
	unsigned short lenB = 0;
	unsigned short lenY;
    unsigned short numArgs = CheckNumArgsInt (2, 3); 
    
    // get number of lists
    // 2 lists = convolution
    // 3 lists = recursive equation
	recEq = numArgs == 3;
	
	TRY	
		// get input signal
		lenX = GetFloatListArg (argptr, &xHdl);		
		
		argptr = next_expression_index (argptr);

		// get impulse response
		lenA = GetFloatListArg (argptr, &aHdl);
								
		// input signal and impulse response is valid
		// (when HeapReallocThrow would have failed we would not have arrived here)
		if (lenX > 0 && 
			lenA > 0) {		
		
			// recursive equation
			if (recEq) {
				argptr = next_expression_index (argptr);
				
				// get recursion coefficients
				lenB = GetFloatListArg (argptr, &bHdl); 
				
				// invalid recursion coefficients
				if (lenB == 0) {
					ER_throwVar (NON_CONFORMING_LISTS_ERROR); 
				}
			}
            
			// length of output signal
			lenY = lenX + lenA - 1;

			// allocate array for output signal
			yHdl = HeapAllocThrow (lenY * sizeof (float)); 

			// dereference them after the last HeapRealloc call
			// (HeapRealloc is called indirectly by GetFloatListArg)
			xPtr = HeapDeref (xHdl);
			aPtr = HeapDeref (aHdl);					
		    
		    if (recEq) {
		    	bPtr = HeapDeref (bHdl);
		    }				

			yPtr = HeapDeref (yHdl);
		
			// bPtr == NULL convolution 
			// bPtr != NULL recursion
			ConRecEq(xPtr,	// input
				lenX, 		// length of input
				aPtr, 		// impulse response
				lenA, 		// length of impulse response
				bPtr, 		// recursion coefs
				lenB, 		// length of recursion coefs
				yPtr); 		// length of output
		
			// push list in reverse order to estack
			push_float_list (yPtr, lenY); 
		} else {
			ER_throwVar (ER_DOMAIN);
		}
	FINALLY
		// HeapFreeIndir handles H_NULL correctly
		HeapFreeIndir ((HANDLE *)&xHdl);
		HeapFreeIndir ((HANDLE *)&aHdl);
		HeapFreeIndir ((HANDLE *)&bHdl);
		HeapFreeIndir ((HANDLE *)&yHdl);
	ENDFINAL
}

/* Multiply and Accumulate
input or output
length of input or output
coef a or coef b
length of coef a or coef b
index in output */
static double MulAcc(double xy[],
	unsigned short lxy,
	double ab[],
	unsigned short lab,
	unsigned short iy)
{
	// input or output index
	short k;
	
	// impulse response or recursion coefficients index
	unsigned short j;
	
	// sum of products
	double s = 0;
	
	// loop over impulse response or recursion coefficients
	for (j = 0; j < lab; j++) {
		k = iy - j;
			
		// valid input or output signal index
		if (0 <= k && k < (short)lxy) {
			s += ab[j] * xy[k]; 
		}
	}
	
	return s;	
}

/* Convolution or Recursive Equation
calculate output signal from input signal using recursion coefficients
input signal
length of input signal 
a coefficients (weight of input signal)
length of a coefficients 
b coefficients (weight of output signal)
length of b coefficients 
output signal */
void ConRecEq(double x[],
	unsigned short xl,
	double a[],
	unsigned short al,
	double b[],
	unsigned short bl,
	double y[])
{  	
	// length of output signal
 	unsigned short yl = xl + al - 1;
 	
 	unsigned short i;
 	 	
	// loop over output array
	for (i = 0; i < yl; i++) {
		// convolve input with impulse response
		y[i] = MulAcc (x,				// input
				xl, 					// length of input
				a, 						// impulse response
				al, 					// length of impulse response
				i) + 					// output index
			// convolve output with recursion coefficients
			(b == NULL ? 0 : MulAcc (y, // output
				yl, 					// length of output
				b, 						// recursion coefs
				bl, 					// length of recursion coefs
				i - 1));				// output index
	}
}

