 /*
		discrete cosine transform (by FFT)
		
    Copyright (C) 2015  Ralph Schenn

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <tiams.h>
#include "Complex.h"
#include "Debug.h"
#include "FFT.h"
#include "Real.h"
#include "Stack.h"

/* odd elements of array augmented with even elements of array 
length must be even 
{1,2,3,4} -> {1,3,4,2} 
forward tranform requires this before FFT, inverse transform after IFFT */
void OddReversedEven(Complex *in,
	unsigned short len,
	Complex *out)
{
	short i, j;

	// loop over odd input elements and
	for (i = 0, j = 0; i < len; i += 2, j++) {
		// transfer them to the output
		out[j] = in[i];
	}		

	// loop over even input elements in reversed order and
	for (i = len - 1; i > 0; i -= 2, j++) {
		// transfer them to the output
		out[j] = in[i];
	}		
}

/* shift position of elements and copy from input to output
length must be even
{1,3,4,2} -> {1,2,3,4} */
void UndoOddReversedEven(Complex *in,
	unsigned short len,
	Complex *out)
{
	short i, j;

	// loop over first half of input elements and
	for (i = 0, j = 0; i < len / 2; i++, j += 2) {
		// transfer them to the output
		out[j] = in[i];
	}		

	// loop over second half of input elements and
	for (i = len / 2, j = len - 1; i < len; i++, j -= 2) {
		// transfer them to the output
		out[j] = in[i];
	}		
}

/* scale elements of array 
forward transform requires this after FFT, inverse transform requires this before IFFT */
void DCTScale(Complex *arr, 
	unsigned short len,
	short dir)
{
    Access_AMS_Global_Variables;

	unsigned short i;
	Complex t;
	double u;

	// t = -i * pi / 2 / len
	t.real = 0.;
	t.imag = -dir;
	t = CMulReal (FLOATTAB[FPI_PIDIV2] / (double)len, t);
	
	u = sqrt (dir == 1 ? 0.5 / (double)len : (double)len / 2.);

	// scale the first element in frequency domain			
	// outPtr[0] *= 2 * sqrt (1 / 4 / len)
	arr[0] = CMulReal ((dir == 1 ? 2. : 1.) * sqrt (dir == 1 ? 0.25 / (double)len : (double)len / 4.), arr[0]);
	
	// loop over elements of transformed input in frequency domain
	for (i = 1; i < len; i++) {
		// scale the other elements in frequency domain
		// outPtr[i] = 2 * real (sqrt (1 / 2 / len) * exp (i * t))        
		arr[i] = CMul (CMulReal (u, CExp (CMulReal (i, t))), arr[i]);
		
		// forward transform
		if (dir == 1) {
			arr[i].real *= 2.;
			arr[i].imag = 0.;
		}
	}
}

#define DIR 1

/* discrete cosine transform (by FFT) 
inspired by dct and idct in octave
(this version is more restrictive allowing only lengths that are a power of two)
push real list 
input must be real */
void dct(void)
{
    Access_AMS_Global_Variables;

	short dir;
	unsigned short i, len;
	volatile HANDLE inHdl, outHdl;
	Complex *inPtr, *outPtr;        
	
	CheckNumArgs (2);
	
	// get second argument
	dir = GetSignShortArgPos (DIR);
	
	// check value of second argument
	if (abs (dir) != 1) {
		ER_throwVar (ER_DOMAIN);
	}

	TRY
		// get input list
		len = GetComplexFloatListArg (top_estack, &inHdl);
		
		// input list is empty or not a power of two
		if (!len || !IsPow2 (len)) {
			ER_throwVar (ER_DOMAIN);
		}
		
		// input list has at least two elements
		if (len > 1) {
			// alloacte space for transformed input list
			outHdl = HeapAllocThrow (len * sizeof (Complex));
	
			inPtr = HeapDeref (inHdl);
			outPtr = HeapDeref (outHdl);
														
			// forward transform
			if (dir == 1) {
				// transform the input list to output list
				OddReversedEven (inPtr, len, outPtr);
				
				FFT (outPtr, len, dir == -1 ? TRUE : FALSE);
								
				DCTScale (outPtr, len, dir);
			// inverse transform
			} else {                  
				DCTScale (inPtr, len, dir);
				
				//push_complex_float_list (inPtr, len);
				//return; 							

				FFT (inPtr, len, dir == -1 ? TRUE : FALSE);
				
				// scale result inverse FFT
				for (i = 0; i < len; i++) {
					inPtr[i].real *= 2.;
					inPtr[i].imag = 0.;
				}
				
				// transform the input list to output list
				UndoOddReversedEven (inPtr, len, outPtr);
			}
				
			// push result of transform
			push_complex_float_list (outPtr, len); 
		// input list has only one element
		} else { 
			push_quantum (END_TAG);
			push_complex_float (((Complex *)HeapDeref (inHdl))[0]);
			push_quantum (LIST_TAG);			
		}
	FINALLY
		HeapFreeIndir ((HANDLE *)&inHdl);
		HeapFreeIndir ((HANDLE *)&outHdl);
	ENDFINAL
}

