/*
		error function
		
    Copyright (C) 2015  Ralph Schenn

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <tiams.h>
#include "Erf.h"
#include "Stack.h"

/* error function
https://en.wikipedia.org/wiki/Error_function#Numerical_approximation */
double Erf(double x)
{
	#define LEN_ERF 10
	const double erfCoefs[] = {
		-1.26551223,
		1.00002368,
		0.37409196,
		0.09678418,
		-0.18628806,
		0.27886807,
		-1.13520398,
		1.48851587,
		-0.82215223,
		0.17087277
	};

	double t = 1 / (1 + 0.5 * fabs(x));

	double tau = 0;
	unsigned char i;

	for (i = 0; i < LEN_ERF; i++) {
		tau += erfCoefs[i] * pow(t, i);
	}

	tau = t * exp(-(x * x) + tau);
	
	return x >= 0 ? 1 - tau : tau - 1;	
}

/* complementary error function */
double ErfC(double x)
{
	return 1.0 - Erf(x); 
}

/* McLaurin series
coefficients
length of coefficient array 
argument of function */
static double McLaurin(const double *coefs, 
	unsigned char length,
	double x)
{
	double s = 0;
    unsigned char i;
    
	for (i = 0; i < length; i++) {
		s += coefs[i] * pow (x, 2 * i + 1);
	}  
	
	return s;
}

/* inverse error function   
may be useable for -1<=x<=1 but not outside this interval
even near the borders of this interval the precision is only around 3 digits
http://mathworld.wolfram.com/InverseErf.html */
double ErfI(double x)
{
	Access_AMS_Global_Variables; 

	#define LEN_ERFI 10
	
	// using only 50% of the coefs reduces execution time by 64%
	const double erfiCoefs[] = {  
		5.0000000000000000E-1,
		1.3089969389957500E-1,
		7.1965865424609900E-2,
		4.8831809751960300E-2,
		3.6649539683190400E-2,
		2.9186250439292300E-2,
		2.4168031585089100E-2,
		2.0573697470262400E-2,
		1.7878606546182300E-2,
		1.5786381659923300E-2
		/*1.4117307386262800E-2,
		1.2756249369960500E-2,
		1.1626130840321400E-2,
		1.0673461112182900E-2*/
	};
		
	return fabs(x) <= 1 ? 1.7724538509055160 * McLaurin(erfiCoefs, 
		LEN_ERFI,
		x) : FLOATTAB[FPI_INVALID];
}

/* push error function */
void push_erf (EStackIndex i)
{
	// push value of function as result
	push_Float (Erf (estack_number_to_Float_inf (i)));
}

/* push inverse error function */
void push_erfi (EStackIndex i)
{
	// push value of function as result
	push_Float (ErfI (estack_number_to_Float_inf (i)));
}
