/*
		Estack Interaction
		
    Copyright (C) 2015  Ralph Schenn with support from Kevin Kofler

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef _STACK_H_
#define _STACK_H_

#include <tiams.h>
#include "Complex.h"

/* convert bool value to corresponding tag value */
#define bool_to_tag(b) ((b) + 43)

// A R G U M E N T   C H E C K I N G //////////////////////////////////////////

/*********************************************************/
/* Title: remaining_element_count */
/* Description: Get number of expressions in tail. */
/* Input: estack index i */
/* i indexes a tail of expressions terminated */
/* by END_TAG. */
/* Output: unsigned number of expressions in the tail */
/*********************************************************/
unsigned short remaining_element_count (EStackIndex i);

/* check if number of arguments is correct */
unsigned char CheckNumArgs(unsigned char expArgs);

/* check if number of arguments is correct (interval) */
unsigned char CheckNumArgsInt(unsigned char minExpArgs, 
	unsigned char maxExpArgs);

/* get type of argument with number (0-based index) */
enum Tags GetArgNumType(short argNum);

/* check if char argument is contained in list
returns the value of the argument at position argPos */
const char *CheckArgCharVal(unsigned short argPos, 
	char *allowedVal);

/* check if tag is boolean */
Boolean IsBool(EStackIndex i);

/* check if tag is integer */
Boolean IsInt(EStackIndex i);

/* check if tag is float */
Boolean IsFloat(EStackIndex i);

/* check if tag is complex float */
Boolean IsCFloat(EStackIndex i);

/* check if tag is string */
Boolean IsString(EStackIndex i);

/* check if tag is true */
Boolean IsTrue(EStackIndex i);

/* all members of list or mat have type 
also works for scalar expressions */
Boolean IsTypeAggregate(EStackIndex argptr,
	Boolean (*funcPtr)(EStackIndex i));

// S C A L A R S //////////////////////////////////////////////////////////////

/* simplify and read short */
signed short estack_to_short_simp (EStackIndex argptr);

/* simplify and read unsigned short */
unsigned short estack_to_ushort_simp (EStackIndex argptr);

/* get signed int from estack */
int GetSignShortArg(EStackIndex *ptr);

/* get signed int argument at position */
signed short GetSignShortArgPos(unsigned char pos);

/* get signed long from estack */
int GetSignLongArg(EStackIndex *ptr);

/* push complex float to estack (16.7) */
void push_complex_float(Complex input);

/* push boolean to stack */
void push_bool(Boolean b);

/* get float from expression stack (allows infinities) */
double estack_number_to_Float_inf(EStackIndex ptr);

/* get int, float or complex from estack 
argument may be int, real or complex */
Complex estack_number_to_Complex(EStackIndex ptr);

/* get complex float from estack 
argument must be complex */
Complex GetComplexFloatArg(EStackIndex *ptr);

// S T R I N G S //////////////////////////////////////////////////////////////

/* get string argument from stack 
caution: returns only pointer to string on estack and not copy of the string 
so we will get errors if overwriting the string on the estack before using 
this pointer 
also works when var conatining string is passed insted of string literal */
const char *GetStrnArg(EStackIndex argptr);

/* get variable name from stack */ 
const char *GetVarArg(EStackIndex argptr);

/* get string argument at position */
const char *GetStrnArgPos(short pos);

/* get string argument and copy it to another variable
(to avoid loosing its content while overwriting the stack) */
void GetStrnArgCopy(EStackIndex ptr, char *dest, 
	unsigned char destLen);

/* push string to estack */
void push_AMS_string(EStackIndex ptr);

// L I S T S //////////////////////////////////////////////////////////////////

/* check if index is list but not matrix */
Boolean is_list(EStackIndex i);

/* count the number of elements in list */
unsigned short GetListDim(EStackIndex ptr);

/* return type of list 
only the first cell is checked */
enum Tags ListType(EStackIndex ptr);

/* push empty list */
void push_empty_list(void);

/* get list of floats 
content is written to output (size of array is adjusted automatically)
returns number of elements */
unsigned short GetFloatListArg(EStackIndex ptr, 
	volatile HANDLE *output);

/* get list of expressions 
content is written to output (size of array is adjusted automatically)
returns number of elements */
unsigned short GetExpressionListArg(EStackIndex ptr,
	volatile HANDLE *output); 

/* push array of floats to estack creating list */
void push_float_list(double input[], 
	short length);
	
/* push array of expressions to estack creating list */
void push_expression_list(EStackIndex input[], 
	short length);

/* push array of floats to estack creating list in reverse order */
void push_float_reverse_list(float input[], 
	unsigned short length);

/* get list of floats 
content is written to output (size of array is adjusted automatically)
returns number of elements */
unsigned short GetComplexFloatListArg(EStackIndex ptr, 
	volatile HANDLE *output);

/* push list of complex floats to estack */
void push_complex_float_list(Complex input[], 
	short length);

/* push array of strings to estack creating list */
void push_string_list_handle(HANDLE input[], 
	short length);

/* push array of expressions to estack creating list */
void push_list(EStackIndex input[], 
	short length);

// M A T R I C E S ////////////////////////////////////////////////////////////

/* return type of matrix 
only the first cell is checked */
enum Tags MatrixType(EStackIndex ptr);

/* count the number of rows and cols in mat */
void GetMatDim(EStackIndex ptr, 
	unsigned short *rows, 
	unsigned short *cols); // ???

/* get matrix of long ints 
content is written to output (size of array is adjusted automatically)
also returns number of rows and colums */
void GetLongintMatArg(EStackIndex *argptr, 
	HANDLE *output, 
	unsigned short *rows,
	unsigned short *cols);

/* push matrix of floats to estack */
void push_longint_mat(long *input, 
	short rows, 
	unsigned short cols);

/* get matrix of floats 
content is written to output (size of array is adjusted automatically)
also returns number of rows and colums */
void GetFloatMatArg(EStackIndex argptr, 
	volatile HANDLE *output, 
	unsigned short *rows,
	unsigned short *cols);

/* get matrix of expressions 
content is written to output (size of array is adjusted automatically)
also returns number of rows and colums
{ { ELEM... } { ELEM... } }
END_TAG END_TAG ELEM... LIST_TAG END_TAG ELEM... LIST_TAG LIST_TAG */
void GetExpressionMatArg(EStackIndex argptr, // 
	volatile HANDLE *output, 
	unsigned short *rows,
	unsigned short *cols);

/* push matrix of floats to estack
array (one-dimensional using explicit pointer arithmetic)
number of rows
number of columns */
void push_float_mat(double *input, 
	short rows, 
	unsigned short cols);
	
/* push matrix of expressions to estack
array (one-dimensional using explicit pointer arithmetic)
number of rows
number of columns */
void push_expression_mat(EStackIndex *input, 
	short rows, 
	unsigned short cols);

/* get matrix of complex floats 
content is written to output (size of array is adjusted automatically)
also returns number of rows and colums
{ { ELEM... } { ELEM... } }
LIST_TAG LIST_TAG ELEM... END_TAG LIST_TAG ELEM... END_TAG END_TAG */
void GetComplexFloatMatArg(EStackIndex argptr, // ok
	volatile HANDLE *output, 
	unsigned short *rows,
	unsigned short *cols);

/* push matrix of complex floats to estack */
void push_complex_float_mat(Complex *input, 
	short rows, 
	unsigned short cols);

// U T I L I T I E S //////////////////////////////////////////////////////////

/* reverse the order of the array elements */
void ReverseArray(EStackIndex array[]);

/* determine begin of stack */
void StackBegin(EStackIndex *ptr);

#endif
